import asyncio
import json
import logging
from typing import Any, Dict, List, Optional

from pyee.asyncio import AsyncIOEventEmitter
from websockets.legacy.client import WebSocketClientProtocol, connect


class GstSignalling(AsyncIOEventEmitter):
    """Signalling peer for the GStreamer WebRTC implementation.

    This class is used to communicate with a GStreamer WebRTC signalling server.
    It relies on an async websocket client to communicate with the server.

    It can be used for the consumer, producer and listener roles. All following messages are supported:

    Peer --> Server
    - "SetPeerStatus": Set current peer status (see set_peer_status)
    - "StartSession":  Start a session with a producer peer (see start_session)
    - "EndSession": End an existing session (see end_session)
    - "Peer": Send a message to a peer the sender is currently in session with (see send_peer_message)
    - "List": Retrieve the current list of producers (see get_list)

    Server --> Peer
    - "Welcome": Welcoming message, sets the Peer ID linked to a new connection
    - "PeerStatusChanged": Notifies listeners that a peer status has changed
    - "StartSession": Instructs a peer to generate an offer and inform about the session ID
    - "SessionStarted": Let consumer know that the requested session is starting with the specified identifier
    - "EndSession": Signals that the session the peer was in was ended
    - "Peer": Messages directly forwarded from one peer to another
    - "List": Provides the current list of consumer peers
    - "Error": Notifies that an error occured with the peer's current session

    Each receive message can be listened to by registering a callback with the corresponding event name.
    For instance, to listen to the "Welcome" message, you can do:

    signalling = GstSignalling(host="localhost", port=8443)

    @signalling.on("Welcome")
    def on_welcome(peer_id: str) -> None:
        print(f"Welcome received, peer_id: {peer_id}")
    """

    def __init__(self, host: str, port: int) -> None:
        """Initializes the signalling peer.

        Args:
            host (str): Hostname of the signalling server.
            port (int): Port of the signalling server."""
        AsyncIOEventEmitter.__init__(self)

        self.logger = logging.getLogger(__name__)

        self.ws: Optional[WebSocketClientProtocol] = None
        self.host = host
        self.port = port

        self.peer_id: Optional[str] = None
        self.handler_task: Optional[asyncio.Task[None]] = None

    async def connect(self) -> None:
        """Connects to the signalling server."""
        if self.ws is not None:
            raise RuntimeError("Already connected.")

        url = f"ws://{self.host}:{self.port}"
        self.logger.info(f"Connecting to {url}")
        self.ws = await connect(url, ping_interval=None)
        self.logger.info("Connected.")

        self.handler_task = asyncio.create_task(self._handler())

    async def close(self) -> None:
        """Closes the connection to the signalling server."""
        if self.ws is None:
            raise RuntimeError("Not connected.")

        if self.handler_task is not None:
            self.handler_task.cancel()
            await self.handler_task
            self.handler_task = None

        self.logger.info("Closing connection.")
        await self.ws.close()
        self.logger.debug("Closed.")

    # Messages (server --> peer)
    async def _handler(self) -> None:
        assert self.ws is not None

        self.logger.info("Starting input message handler.")

        try:
            async for data in self.ws:
                assert isinstance(data, str)

                self.logger.debug(f"Received message: {data}")
                message: Dict[str, Any] = json.loads(data)
                await self._handle_messages(message)
        except asyncio.CancelledError:
            self.logger.info("Input message handler cancelled.")

    async def _handle_messages(self, message: Dict[str, Any]) -> None:
        # Welcoming message, sets the Peer ID linked to a new connection
        if message["type"] == "welcome":
            peer_id = message["peerId"]
            self.peer_id = peer_id
            self.emit("Welcome", peer_id)

        # Notifies listeners that a peer status has changed
        elif message["type"] == "peerStatusChanged":
            roles = message["roles"]
            meta = message["meta"]
            peer_id = message["peerId"]

            self.emit("PeerStatusChanged", peer_id, roles, meta)

        # Instructs a peer to generate an offer and inform about the session ID
        elif message["type"] == "startSession":
            peer_id = message["peerId"]
            session_id = message["sessionId"]
            self.emit("StartSession", peer_id, session_id)

        # Let consumer know that the requested session is starting with the specified identifier
        elif message["type"] == "sessionStarted":
            peer_id = message["peerId"]
            session_id = message["sessionId"]
            self.emit("SessionStarted", peer_id, session_id)

        # Signals that the session the peer was in was ended
        elif message["type"] == "endSession":
            session_id = message["sessionId"]
            self.emit("EndSession", session_id)

        # Messages directly forwarded from one peer to another
        elif message["type"] == "peer":
            session_id = message["sessionId"]
            message = dict(message)
            del message["type"]
            del message["sessionId"]
            self.emit("Peer", session_id, message)

        # Provides the current list of consumer peers
        elif message["type"] == "list":
            producers = message["producers"]
            producers = {p["id"]: p["meta"] for p in producers}
            self.emit("List", producers)

        # Notifies that an error occured with the peer's current session
        elif message["type"] == "error":
            details = message["details"]
            self.logger.error(f'An error occured: "{details}"')
            self.emit("Error", details)

        else:
            self.logger.warning(f"Received unknown message type: {message}.")

    # Messages (peer --> server)
    async def set_peer_status(self, roles: List[str], name: str) -> None:
        """Sets the peer status.

        Args:
            roles (List[str]): List of roles the peer has (producer, listener).
            name (str): Name of the peer.
        """
        if self.peer_id is None:
            raise RuntimeError("PeerId not yet received.")

        for role in roles:
            if role not in ("listener", "producer"):
                raise ValueError(f"Invalid role {role}.")

        message = {
            "type": "setPeerStatus",
            "roles": roles,
            "meta": {"name": name},
            "peerId": self.peer_id,
        }

        await self._send(message)

    async def start_session(self, peer_id: str) -> None:
        """Starts a session with a producer peer.

        Args:
            peer_id (str): Peer ID of the producer peer.
        """
        if self.peer_id is None:
            raise RuntimeError("PeerId not yet received.")

        message = {"type": "startSession", "peerId": peer_id}
        await self._send(message)

    async def end_session(self, session_id: str) -> None:
        """Ends an existing session.

        Args:
            session_id (str): Session ID.
        """
        message = {"type": "endSession", "sessionId": session_id}
        await self._send(message)

    async def send_peer_message(self, session_id: str, type: str, peer_message: Dict[str, Any]) -> None:
        """Sends a message to a peer the sender is currently in session with.

        Args:
            session_id (str): Session ID.
            type (str): Type of the message (sdp or ice).
            peer_message (str): Message to send (sdp or icecandidate).
        """
        message = {
            "type": "peer",
            "sessionId": session_id,
            type: peer_message,
        }

        await self._send(message)

    async def send_list(self) -> None:
        """Requests the current list of producers."""
        message = {"type": "list"}
        await self._send(message)

    async def _send(self, message: Dict[str, Any]) -> None:
        if self.ws is None:
            raise RuntimeError("Not connected.")

        self.logger.debug(f"Sending message: {message}")
        await self.ws.send(json.dumps(message))



import asyncio
from typing import Dict, List

from .gst_abstract_role import GstSignallingAbstractRole


class GstSignallingListener(GstSignallingAbstractRole):
    def __init__(self, host: str, port: int, name: str) -> None:
        GstSignallingAbstractRole.__init__(self, host=host, port=port)
        self.name = name

        @self.signalling.on("PeerStatusChanged")  # type: ignore[arg-type]
        def on_peer_status_changed(
            peer_id: str,
            roles: List[str],
            meta: Dict[str, str],
        ) -> None:
            self.emit("PeerStatusChanged", peer_id, roles, meta)

    async def connect(self) -> None:
        await super().connect()
        await self.signalling.set_peer_status(roles=["listener"], name=self.name)

    async def serve4ever(self) -> None:
        await self.connect()
        await self.consume()

    async def consume(self) -> None:
        while True:
            await asyncio.sleep(1)

    async def close(self) -> None:
        await self.signalling.close()



import asyncio
import logging
from typing import Any, Dict, NamedTuple, Optional

import gi
from pyee.asyncio import AsyncIOEventEmitter

from .gst_signalling import GstSignalling

gi.require_version("Gst", "1.0")
gi.require_version("GstWebRTC", "1.0")

from gi.repository import Gst  # noqa : E402

GstSession = NamedTuple(
    "GstSession",
    [
        ("peer_id", str),
        ("session_id", str),
        ("pc", Gst.Element),  # type '__gi__.GstWebRTCBin'
    ],
)


class GstSignallingAbstractRole(AsyncIOEventEmitter):
    def __init__(
        self,
        host: str,
        port: int,
    ) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        signalling = GstSignalling(host=host, port=port)

        self.peer_id: Optional[str] = None
        self.peer_id_evt = asyncio.Event()
        self.consume_evt = asyncio.Event()
        self._asyncloop = asyncio.get_event_loop()

        self.sessions: Dict[str, GstSession] = {}

        @signalling.on("Welcome")  # type: ignore[arg-type]
        def on_welcome(peer_id: str) -> None:
            self.peer_id = peer_id
            self.peer_id_evt.set()

        @signalling.on("StartSession")  # type: ignore[arg-type]
        async def on_start_session(peer_id: str, session_id: str) -> None:
            self.logger.info(f"StartSession received, session_id: {session_id}")
            await self.setup_session(session_id, peer_id)

        @signalling.on("SessionStarted")  # type: ignore[arg-type]
        async def on_session_started(peer_id: str, session_id: str) -> None:
            self.logger.info(f"SessionStarted received, session_id: {session_id}")
            await self.setup_session(session_id, peer_id)

        @signalling.on("Peer")  # type: ignore[arg-type]
        async def on_peer(session_id: str, message: Dict[str, Dict[str, Any]]) -> None:
            self.logger.info(f"Peer received, session_id: {session_id}, message: {message}")
            await self.peer_for_session(session_id, message)

        @signalling.on("EndSession")  # type: ignore[arg-type]
        async def on_end_session(session_id: str) -> None:
            self.logger.info(f"EndSession received, session_id: {session_id}")
            await self.close_session(session_id)

        self.signalling = signalling

        Gst.init(None)

        self._pipeline = Gst.Pipeline.new()
        # pipeline will only contain dynamically added webrtcbins
        self._pipeline.set_state(Gst.State.PLAYING)

    def __del__(self) -> None:
        self._pipeline.set_state(Gst.State.NULL)
        # Gst.deinit()

    def make_send_sdp(self, sdp: Any, type: str, session_id: str) -> None:  # sdp is GstWebRTC.WebRTCSessionDescription
        text = sdp.sdp.as_text()
        msg = {"type": type, "sdp": text}
        asyncio.run_coroutine_threadsafe(self.send_sdp(session_id, msg), self._asyncloop)

    def send_ice_candidate_message(self, _: Gst.Element, mlineindex: int, candidate: str, session_id: str) -> None:
        icemsg = {"candidate": candidate, "sdpMLineIndex": mlineindex}
        asyncio.run_coroutine_threadsafe(self.send_ice(session_id, icemsg), self._asyncloop)

    def init_webrtc(self, session_id: str) -> Gst.Element:
        webrtc = Gst.ElementFactory.make("webrtcbin")
        assert webrtc

        webrtc.set_property("bundle-policy", "max-bundle")
        webrtc.connect("on-ice-candidate", self.send_ice_candidate_message, session_id)

        self._pipeline.add(webrtc)

        return webrtc

    async def connect(self) -> None:
        assert self.signalling is not None

        await self.signalling.connect()
        await self.peer_id_evt.wait()

    async def close(self) -> None:
        await self.signalling.close()
        self.consume_evt.set()

    async def consume(self) -> None:
        # while True:
        #    await asyncio.sleep(1000)
        await self.consume_evt.wait()

    # Session management
    async def setup_session(self, session_id: str, peer_id: str) -> GstSession:
        self.logger.info("setup session")
        pc = self.init_webrtc(session_id)

        session = GstSession(peer_id, session_id, pc)

        self.sessions[session_id] = session

        return session

    async def peer_for_session(self, session_id: str, message: Dict[str, Dict[str, str]]) -> None:
        self.logger.info(f"peer for session {session_id} {message}")

    def handle_ice_message(self, webrtc: Gst.Element, ice_msg: Dict[str, Any]) -> None:
        candidate = ice_msg["candidate"]
        sdpmlineindex = ice_msg["sdpMLineIndex"]
        webrtc.emit("add-ice-candidate", sdpmlineindex, candidate)

    async def close_session(self, session_id: str) -> None:
        self.logger.info("close session")
        try:
            session = self.sessions.pop(session_id)
            self._pipeline.remove(session.pc)
            session.pc.set_state(Gst.State.NULL)
            await self.signalling.end_session(session_id)
            # self.emit("close_session", session)
        except KeyError:
            self.logger.warning("Session not found")

    async def send_sdp(self, session_id: str, sdp: Dict[str, Dict[str, str]]) -> None:
        await self.signalling.send_peer_message(session_id, "sdp", sdp)

    async def send_ice(self, session_id: str, ice: Dict[str, Any]) -> None:
        await self.signalling.send_peer_message(session_id, "ice", ice)



import argparse
import asyncio
from typing import Dict

from .gst_signalling import GstSignalling


async def get_list(host: str, port: int) -> Dict[str, Dict[str, str]]:
    producers = {}
    got_it = asyncio.Event()

    signalling = GstSignalling(host=host, port=port)
    await signalling.connect()

    @signalling.on("List")  # type: ignore[arg-type]
    def on_list(found_producers: Dict[str, Dict[str, str]]) -> None:
        producers.update(found_producers)
        got_it.set()

    await signalling.send_list()
    await got_it.wait()

    await signalling.close()

    return producers


def get_producer_list(host: str, port: int) -> Dict[str, Dict[str, str]]:
    """Gets the list of producers from the signalling server.

    Args:
        host (str): Hostname of the signalling server.
        port (str): Port of the signalling server.
    Returns:
        Dict[str, Dict[str, str]]: Dictionary of producers, where the key is the producer ID
        and the value is a dictionary with the producer metadata (eg. name).
    """

    loop = asyncio.get_event_loop()
    return loop.run_until_complete(get_list(host, port))


def find_producer_peer_id_by_name(host: str, port: int, name: str) -> str:
    """Finds the peer ID of a producer by its name.

    Args:
        name (str): Name of the producer.
    Returns:
        str: Producer peer ID (if multiple entries, the first is returned).
    Raises:
        KeyError: If the producer is not found.
    """
    producers = get_producer_list(host=host, port=port)

    for producer_id, producer_meta in producers.items():
        if producer_meta["name"] == name:
            return producer_id

    raise KeyError(f"Producer {name} not found.")


async def async_find_producer_peer_id_by_name(host: str, port: int, name: str) -> str:
    """Finds the peer ID of a producer by its name.

    Args:
        name (str): Name of the producer.
    Returns:
        str: Producer peer ID (if multiple entries, the first is returned).
    Raises:
        KeyError: If the producer is not found.
    """
    producers = await get_list(host=host, port=port)

    for producer_id, producer_meta in producers.items():
        if producer_meta["name"] == name:
            return producer_id

    raise KeyError(f"Producer {name} not found.")


def add_signaling_arguments(parser: argparse.ArgumentParser) -> None:
    """Adds command line arguments for GstSignaling.

    * signaling-host: Hostname of the signalling server.
    * signaling-port: Port of the signalling server.
    * role: Signalling role (consumer or producer).
    * name: Peer name.
    * remote-producer-peer-id: Producer peer_id (required in consumer role!).
    """
    parser.add_argument("--signaling-host", default="127.0.0.1", help="Gstreamer signaling host")
    parser.add_argument("--signaling-port", default=8443, help="Gstreamer signaling port")
    parser.add_argument("role", choices=["consumer", "producer"], help="Signaling role")
    parser.add_argument("--name", default="my-name", help="peer name")
    parser.add_argument(
        "--remote-producer-peer-id",
        type=str,
        help="producer peer_id (in consumer role, either set this or remote-producer-peer-name!)",
    )
    parser.add_argument(
        "--remote-producer-peer-name",
        type=str,
        default="remote-producer-peer-name",
        help="producer peer_name (in consumer role, either set this or remote-producer-peer-id!)",
    )



import argparse
import asyncio
import logging
from typing import Dict, List

from gst_signalling import GstSignallingListener


def main(args: argparse.Namespace) -> None:
    listener = GstSignallingListener(
        host=args.signaling_host,
        port=args.signaling_port,
        name=args.name,
    )

    @listener.on("PeerStatusChanged")  # type: ignore[misc]
    def on_peer_status_changed(
        peer_id: str,
        roles: List[str],
        meta: Dict[str, str],
    ) -> None:
        print(f'Peer "{peer_id}" changed roles to {roles} with meta {meta}')

    # run event loop
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(listener.serve4ever())
    except KeyboardInterrupt:
        pass
    finally:
        loop.run_until_complete(listener.close())


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--signaling-host", default="127.0.0.1", help="Gstreamer signaling host")
    parser.add_argument("--signaling-port", default=8443, help="Gstreamer signaling port")
    parser.add_argument("--name", default="listener", help="Producer name")
    parser.add_argument("--verbose", "-v", action="count", default=0)
    args = parser.parse_args()

    if args.verbose == 1:
        logging.basicConfig(level=logging.INFO)
    elif args.verbose > 1:
        logging.basicConfig(level=logging.DEBUG)

    main(args)



import logging
from typing import Dict

from gi.repository import Gst, GstSdp, GstWebRTC

from .gst_abstract_role import GstSession, GstSignallingAbstractRole


class GstSignallingProducer(GstSignallingAbstractRole):
    def __init__(self, host: str, port: int, name: str) -> None:
        super().__init__(host, port)
        self.name = name
        self.logger = logging.getLogger(__name__)

    async def connect(self) -> None:
        await super().connect()
        await self.signalling.set_peer_status(roles=["producer"], name=self.name)

    async def serve4ever(self) -> None:
        await self.connect()
        await self.consume()

    def on_offer_created(self, promise: Gst.Promise, webrtc: Gst.Element, session_id: str) -> None:
        self.logger.debug(f"on offer created {promise} {webrtc} {session_id}")
        assert promise.wait() == Gst.PromiseResult.REPLIED
        reply = promise.get_reply()
        offer = reply.get_value("offer")  # type: ignore[union-attr]

        promise = Gst.Promise.new()
        self.logger.info("Offer created, setting local description")
        webrtc.emit("set-local-description", offer, promise)
        promise.interrupt()
        self.make_send_sdp(offer, "offer", session_id)

    def on_negotiation_needed(self, element: Gst.Element, session_id: str) -> None:
        self.logger.debug(f"on negociation needed {element} {session_id}")
        promise = Gst.Promise.new_with_change_func(self.on_offer_created, element, session_id)
        element.emit("create-offer", None, promise)

    async def setup_session(self, session_id: str, peer_id: str) -> GstSession:
        session = await super().setup_session(session_id, peer_id)
        self.logger.info("setup session producer")

        # send offer
        pc = session.pc
        pc.connect("on-negotiation-needed", self.on_negotiation_needed, session_id)

        pc.sync_state_with_parent()
        self.emit("new_session", session)

        return session

    async def peer_for_session(self, session_id: str, message: Dict[str, Dict[str, str]]) -> None:
        self.logger.info(f"peer for session {session_id} {message}")

        session = self.sessions[session_id]
        webrtc = session.pc

        if "sdp" in message:
            if message["sdp"]["type"] == "answer":
                self.logger.debug("set remote desc")
                _, sdpmsg = GstSdp.SDPMessage.new_from_text(message["sdp"]["sdp"])

                sdp_type = GstWebRTC.WebRTCSDPType.ANSWER
                answer = GstWebRTC.WebRTCSessionDescription.new(sdp_type, sdpmsg)

                promise = Gst.Promise.new()
                webrtc.emit("set-remote-description", answer, promise)
                promise.interrupt()
                self.logger.debug("set remote desc done")
            elif message["sdp"]["type"] == "offer":
                self.logger.warning("producer should not receive the offer")
            else:
                self.logger.error(f"SDP not properly formatted {message['sdp']}")
        elif "ice" in message:
            self.handle_ice_message(webrtc, message["ice"])
        else:
            self.logger.error(f"message not processed {message}")



import logging
from typing import Dict

import gi

gi.require_version("Gst", "1.0")
gi.require_version("GstWebRTC", "1.0")
gi.require_version("GstSdp", "1.0")

from gi.repository import Gst, GstSdp, GstWebRTC  # noqa : E402

from .gst_abstract_role import GstSession, GstSignallingAbstractRole  # noqa : E402


class GstSignallingConsumer(GstSignallingAbstractRole):
    def __init__(
        self,
        host: str,
        port: int,
        producer_peer_id: str,
    ) -> None:
        super().__init__(host, port)
        self.logger = logging.getLogger(__name__)
        self.producer_peer_id = producer_peer_id

    async def connect(self) -> None:
        await super().connect()
        await self.signalling.start_session(self.producer_peer_id)
        self.logger.info("connect")

    async def setup_session(self, session_id: str, peer_id: str) -> GstSession:
        session = await super().setup_session(session_id, peer_id)
        self.logger.info("setup session consumer")

        self._pipeline.set_state(Gst.State.PLAYING)
        self.emit("new_session", session)

        return session

    def on_answer_created(self, promise: Gst.Promise, webrtc: Gst.Element, session_id: str) -> None:
        assert promise.wait() == Gst.PromiseResult.REPLIED
        reply = promise.get_reply()
        # answer = reply["answer"]
        answer = reply.get_value("answer")  # type: ignore[union-attr]
        promise = Gst.Promise.new()
        webrtc.emit("set-local-description", answer, promise)
        promise.interrupt()  # we don't care about the result, discard it
        self.make_send_sdp(answer, "answer", session_id)

    def on_offer_set(self, promise: Gst.Promise, webrtc: Gst.Element, session_id: str) -> None:
        assert promise.wait() == Gst.PromiseResult.REPLIED
        promise = Gst.Promise.new_with_change_func(self.on_answer_created, webrtc, session_id)
        webrtc.emit("create-answer", None, promise)

    async def peer_for_session(self, session_id: str, message: Dict[str, Dict[str, str]]) -> None:
        self.logger.info(f"peer for session {session_id} {message}")

        session = self.sessions[session_id]
        webrtc = session.pc

        if "sdp" in message:
            if message["sdp"]["type"] == "offer":
                _, sdpmsg = GstSdp.SDPMessage.new_from_text(message["sdp"]["sdp"])
                sdp_type = GstWebRTC.WebRTCSDPType.OFFER
                offer = GstWebRTC.WebRTCSessionDescription.new(sdp_type, sdpmsg)
                promise = Gst.Promise.new_with_change_func(self.on_offer_set, webrtc, session_id)
                webrtc.emit("set-remote-description", offer, promise)
                self.logger.debug("set remote desc done")

            elif message["sdp"]["type"] == "answer":
                self.logger.warning("Consumer should not receive the answer")
            else:
                self.logger.error(f"SDP not properly formatted {message['sdp']}")

        elif "ice" in message:
            self.handle_ice_message(webrtc, message["ice"])

        else:
            self.logger.error(f"message not processed {message}")



import argparse
import os
import subprocess
from typing import Optional

import gi

gi.require_version("Gst", "1.0")
from gi.repository import Gst  # noqa : E402

from gst_signalling.utils import find_producer_peer_id_by_name  # noqa : E402


class GstRecorder:
    def __init__(
        self, signalling_host: str, signalling_port: int, peer_id: Optional[str] = None, peer_name: Optional[str] = None
    ) -> None:
        Gst.init(None)

        self.pipeline = Gst.Pipeline.new("webRTC-recorder")
        self.source = Gst.ElementFactory.make("webrtcsrc")

        if not self.pipeline:
            print("Pipeline could be created.")
            exit(-1)

        if not self.source:
            print(
                "webrtcsrc component could not be created. Please make sure that the plugin is installed \
                (see https://gitlab.freedesktop.org/gstreamer/gst-plugins-rs/-/tree/main/net/webrtc)"
            )
            exit(-1)

        self.pipeline.add(self.source)

        if peer_id is None:
            peer_id = find_producer_peer_id_by_name(signalling_host, signalling_port, peer_name)
            print(f"found peer id: {peer_id}")

        self.source.connect("pad-added", self.webrtcsrc_pad_added_cb)
        signaller = self.source.get_property("signaller")
        signaller.set_property("producer-peer-id", peer_id)
        signaller.set_property("uri", f"ws://{signalling_host}:{signalling_port}")

    def webrtcsrc_pad_added_cb(self, webrtcsrc: Gst.Element, pad: Gst.Pad) -> None:
        if pad.get_name().startswith("video"):  # type: ignore[union-attr]
            videodepay = Gst.ElementFactory.make("rtph264depay")
            assert videodepay is not None
            gdppay = Gst.ElementFactory.make("gdppay")
            assert gdppay is not None
            filesink = Gst.ElementFactory.make("filesink")
            assert filesink is not None
            filesink.set_property("location", f"{pad.get_name()}.gdp")

            self.pipeline.add(videodepay)
            self.pipeline.add(gdppay)
            self.pipeline.add(filesink)
            videodepay.link(gdppay)
            gdppay.link(filesink)
            pad.link(videodepay.get_static_pad("sink"))  # type: ignore[arg-type]

            videodepay.sync_state_with_parent()
            gdppay.sync_state_with_parent()
            filesink.sync_state_with_parent()
        elif pad.get_name().startswith("audio"):  # type: ignore[union-attr]
            audiodepay = Gst.ElementFactory.make("rtpopusdepay")
            assert audiodepay is not None
            gdppay = Gst.ElementFactory.make("gdppay")
            assert gdppay is not None
            filesink = Gst.ElementFactory.make("filesink")
            assert filesink is not None
            filesink.set_property("location", f"{pad.get_name()}.gdp")

            self.pipeline.add(audiodepay)
            self.pipeline.add(gdppay)
            self.pipeline.add(filesink)
            audiodepay.link(gdppay)
            gdppay.link(filesink)
            pad.link(audiodepay.get_static_pad("sink"))  # type: ignore[arg-type]

            audiodepay.sync_state_with_parent()
            gdppay.sync_state_with_parent()
            filesink.sync_state_with_parent()

    def __del__(self) -> None:
        Gst.deinit()

    def get_bus(self) -> Gst.Bus:
        return self.pipeline.get_bus()

    def record(self) -> None:
        # Start playing
        ret = self.pipeline.set_state(Gst.State.PLAYING)
        if ret == Gst.StateChangeReturn.FAILURE:
            print("Error starting playback.")
            exit(-1)
        print("recording ... (ctrl+c to quit)")

    def stop(self) -> None:
        print("stopping")
        self.pipeline.send_event(Gst.Event.new_eos())
        self.pipeline.set_state(Gst.State.NULL)


def process_msg(bus: Gst.Bus) -> bool:
    msg = bus.timed_pop_filtered(10 * Gst.MSECOND, Gst.MessageType.ANY)
    if msg:
        if msg.type == Gst.MessageType.ERROR:
            err, debug = msg.parse_error()
            print(f"Error: {err}, {debug}")
            return False
        elif msg.type == Gst.MessageType.EOS:
            print("End-Of-Stream reached.")
            return False
        # else:
        #    print(f"Message: {msg.type}")
    return True


def save_file(file_name: str) -> None:
    """
    mux the gdp streams with a command similar to
    gst-launch-1.0 \
        mp4mux name=mux ! filesink location=recording.mp4 \
        filesrc location=video_0.gdp ! gdpdepay ! h264parse ! queue ! mux. \
        filesrc location=video_1.gdp ! gdpdepay ! h264parse ! queue ! mux. \
        filesrc location=audio_0.gdp ! gdpdepay ! opusparse ! queue ! mux.
    """

    base_command = f"gst-launch-1.0 mp4mux name=mux ! filesink location={file_name} "
    files = [f for f in os.listdir() if f.endswith(".gdp")]

    if len(files) == 0:
        print("no gdp file found")
        return
    else:
        print(f"gdp files to combine: {files}")

    file_commands = []

    for file in files:
        if os.path.isfile(file):
            if "video" in file:
                file_commands.append(f"filesrc location={file} ! gdpdepay ! h264parse ! queue ! mux.")
            elif "audio" in file:
                file_commands.append(f"filesrc location={file} ! gdpdepay ! opusparse ! queue ! mux.")

    command = base_command + " ".join(file_commands)

    print(f"running : {command}")

    subprocess.run(command, shell=True)


def main() -> None:
    parser = argparse.ArgumentParser(description="webrtc gstreamer simple recorder")
    parser.add_argument("--signaling-host", default="127.0.0.1", help="Gstreamer signaling host")
    parser.add_argument("--signaling-port", default=8443, help="Gstreamer signaling port")
    parser.add_argument(
        "--remote-producer-peer-id",
        type=str,
        help="producer peer_id",
    )
    parser.add_argument(
        "--remote-producer-peer-name",
        type=str,
        help="producer name",
    )

    parser.add_argument("--output", type=str, help="output mp4 file", default="recording.mp4")

    args = parser.parse_args()

    if args.remote_producer_peer_id is None and args.remote_producer_peer_name is None:
        exit("You must set either remote_producer_peer_id or remote_producer_peer_name")

    recorder = GstRecorder(
        args.signaling_host, args.signaling_port, args.remote_producer_peer_id, args.remote_producer_peer_name
    )
    recorder.record()

    # Wait until error or EOS
    bus = recorder.get_bus()
    try:
        while True:
            if not process_msg(bus):
                break

    except KeyboardInterrupt:
        print("User exit")
    finally:
        # Free resources
        recorder.stop()

    save_file(args.output)


if __name__ == "__main__":
    main()



# from .gst_abstract_role import GstSession  # noqa: F401
from .gst_consumer import GstSignallingConsumer  # noqa: F401
from .gst_listener import GstSignallingListener  # noqa: F401
from .gst_producer import GstSignallingProducer  # noqa: F401



import argparse
import logging

from gst_signalling import utils


def get_producer_list(args: argparse.Namespace) -> None:
    """Main function."""
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    producers = utils.get_producer_list(args.signalling_host, args.signalling_port)

    if producers:
        print("List received, producers:")
        for producer_id, producer_meta in producers.items():
            print(f"  - {producer_id}: {producer_meta}")
    else:
        print("List received, no producers.")


def main() -> None:
    parser = argparse.ArgumentParser(description="Get gstreamer producer list")
    parser.add_argument("--signalling-host", default="127.0.0.1")
    parser.add_argument("--signalling-port", default=8443, type=int)
    parser.add_argument("--verbose", "-v", action="count")
    args = parser.parse_args()

    get_producer_list(args)


if __name__ == "__main__":
    main()
